#pragma once

#define POSED_LIST_SIZE 12

void posEdScrollDown(void);
void posEdScrollUp(void);
void posEdScrollToTop(void);
void posEdScrollToBottom(void);
void posEdPageUp(void);
void posEdPageDown(void);
void posEdClearNames(void);
void posEdEditName(void);
void posEdInsert(void);
void posEdDelete(void);
void posEdToggle(void);
void renderPosEdScreen(void);
void updatePosEd(void);
