#include <winver.h>
#include "pt2_header.h" // PROG_VER_STR

VS_VERSION_INFO VERSIONINFO
FILEVERSION     1,0,0,0
FILETYPE        VFT_APP

BEGIN
	BLOCK "StringFileInfo"
	BEGIN
		BLOCK "040904b0"
		BEGIN
			VALUE "FileDescription",  "ProTracker 2 clone for Windows"
			VALUE "ProductName",      "ProTracker 2 clone"
			VALUE "ProductVersion",   PROG_VER_STR
#ifdef _WIN64
			VALUE "InternalName",     "pt2-clone-win64"
			VALUE "OriginalFilename", "pt2-clone-win64.exe"
#else
			VALUE "InternalName",     "pt2-clone-win32"
			VALUE "OriginalFilename", "pt2-clone-win32.exe"
#endif
			VALUE "LegalCopyright",   "Copyright © Olav ""8bitbubsy"" Sørensen"
		END
	END

	BLOCK "VarFileInfo"
	BEGIN
		VALUE "Translation", 0x0409, 1200
	END
END

IDI_MAIN_ICON ICON "gfx\\pt2-clone.ico"
